/*
   DDS, a bridge double dummy solver.

   Copyright (C) 2006-2014 by Bo Haglund /
   2014-2018 by Bo Haglund & Soren Hein.

   See LICENSE and README.
*/

#ifndef DTEST_TESTCOMMON_H
#define DTEST_TESTCOMMON_H

int realMain(int argc, char * argv[]);

#endif

